#include <linux/hid.h>
#include <linux/module.h>
#include <linux/usb.h>
#include "usbhid/usbhid.h"
#include <linux/init.h>

//update FW
#include <linux/fs.h>
#include <linux/cdev.h>
#include <asm/uaccess.h>	//copy_from_user() & copy_to_user()

#include "hid-ids.h"
#include "hid-sis_ctrl.h"

static int sis_char_devs_count = 1;        /* device count */
static int sis_char_major = 0;
static struct cdev sis_char_cdev;
static struct class *sis_char_class = NULL;

static struct hid_device *hid_dev_backup = NULL;  //backup address
static struct urb *backup_urb = NULL;

int sis_cdev_open(struct inode *inode, struct file *filp)	//20120306 Yuger ioctl for tool
{
	struct usbhid_device *usbhid; 

	//20110511, Yuger, kill current urb by method of usbhid_stop
	if ( !hid_dev_backup )
	{
		printk( KERN_INFO "(stop)hid_dev_backup is not initialized yet" );
		return -1;
	}

	usbhid = hid_dev_backup->driver_data;

	printk( KERN_INFO "sys_sis_HID_stop\n" );

	//printk( KERN_INFO "hid_dev_backup->vendor, hid_dev_backup->product = %x %x\n", hid_dev_backup->vendor, hid_dev_backup->product );

	//20110602, Yuger, fix bug: not contact usb cause kernel panic
	if( !usbhid )
	{
		printk( KERN_INFO "(stop)usbhid is not initialized yet" );
		return -1;
	}
	else if ( !usbhid->urbin )
	{
		printk( KERN_INFO "(stop)usbhid->urbin is not initialized yet" );
		return -1;
	}
	else if (hid_dev_backup->vendor == USB_VENDOR_ID_SIS2_TOUCH)
	{
		usb_fill_int_urb(backup_urb, usbhid->urbin->dev, usbhid->urbin->pipe,
			usbhid->urbin->transfer_buffer, usbhid->urbin->transfer_buffer_length,
			usbhid->urbin->complete, usbhid->urbin->context, usbhid->urbin->interval);

                clear_bit( HID_STARTED, &usbhid->iofl );
                set_bit( HID_DISCONNECTED, &usbhid->iofl );

                usb_kill_urb( usbhid->urbin );
                usb_free_urb( usbhid->urbin );
                usbhid->urbin = NULL;
		return 0;
	}
        else	
	{
		printk (KERN_INFO "This is not a SiS device");
		return -801;
	}
}

int sis_cdev_release(struct inode *inode, struct file *filp)
{
	//20110505, Yuger, rebuild the urb which is at the same urb address, then re-submit it

	int ret;
	struct usbhid_device *usbhid;
	unsigned long flags;
	
	if ( !hid_dev_backup )
	{
		printk( KERN_INFO "(stop)hid_dev_backup is not initialized yet" );
		return -1;
	}

	usbhid = hid_dev_backup->driver_data;

	printk( KERN_INFO "sys_sis_HID_start" );

	if( !usbhid )
	{
		printk( KERN_INFO "(start)usbhid is not initialized yet" );
		return -1;
	}

	if( !backup_urb )
	{
		printk( KERN_INFO "(start)backup_urb is not initialized yet" );
		return -1;
	}

	clear_bit( HID_DISCONNECTED, &usbhid->iofl );
	usbhid->urbin = usb_alloc_urb( 0, GFP_KERNEL );

	if( !backup_urb->interval )
	{
		printk( KERN_INFO "(start)backup_urb->interval does not exist" );
		return -1;
	}

	usb_fill_int_urb(usbhid->urbin, backup_urb->dev, backup_urb->pipe, 
		backup_urb->transfer_buffer, backup_urb->transfer_buffer_length, 
		backup_urb->complete, backup_urb->context, backup_urb->interval);
	usbhid->urbin->transfer_dma = usbhid->inbuf_dma;
	usbhid->urbin->transfer_flags |= URB_NO_TRANSFER_DMA_MAP;

	set_bit( HID_STARTED, &usbhid->iofl );

	//method at hid_start_in
	spin_lock_irqsave( &usbhid->lock, flags );		
	ret = usb_submit_urb( usbhid->urbin, GFP_ATOMIC );
	spin_unlock_irqrestore( &usbhid->lock, flags );
	//yy

//	printk( KERN_INFO "ret = %d", ret );

	return ret;
}

//SiS 817 only
ssize_t sis_cdev_read(struct file *file, char __user *buf, size_t count, loff_t *ppos)
{
	int actual_length = 0, timeout = 0;
	u8 *rep_data = NULL;
	u16 size = 0;
	long rep_ret;
	struct usb_interface *intf = to_usb_interface(hid_dev_backup->dev.parent);
	struct usb_device *dev = interface_to_usbdev(intf);

	size = (((u16)(buf[64] & 0xff)) << 24) + (((u16)(buf[65] & 0xff)) << 16) + 
		(((u16)(buf[66] & 0xff)) << 8) + (u16)(buf[67] & 0xff);
	timeout = (((int)(buf[68] & 0xff)) << 24) + (((int)(buf[69] & 0xff)) << 16) + 
		(((int)(buf[70] & 0xff)) << 8) + (int)(buf[71] & 0xff);

	rep_data = kzalloc(size, GFP_KERNEL);
	if (!rep_data)
		return -12;

	if ( copy_from_user( rep_data, (void*)buf, size) ) 
	{
		printk( KERN_INFO "copy_to_user fail\n" );
		//free allocated data
		kfree( rep_data );
		rep_data = NULL;
		return -19;
	}

	rep_ret = usb_interrupt_msg(dev, backup_urb->pipe,
		rep_data, size, &actual_length, timeout);

	if( rep_ret == 0 )
	{
		rep_ret = actual_length;
		if ( copy_to_user( (void*)buf, rep_data, rep_ret ) ) 
		{
			printk( KERN_INFO "copy_to_user fail\n" );
			//free allocated data
			kfree( rep_data );
			rep_data = NULL;
			return -19;
		}
	}

	//free allocated data
	kfree( rep_data );
	rep_data = NULL;
	//printk( KERN_INFO "sis_cdev_read : rep_ret = %d\n", rep_ret );
	return rep_ret;
}

ssize_t sis_cdev_write( struct file *file, const char __user *buf, size_t count, loff_t *f_pos )
{
	int i, actual_length = 0;
	u8 *tmp_data = NULL;	//include report id
	u8 *rep_data = NULL;
	long rep_ret;
	struct usb_interface *intf = to_usb_interface( hid_dev_backup->dev.parent );
	struct usb_device *dev = interface_to_usbdev( intf );
	struct usbhid_device *usbhid = hid_dev_backup->driver_data;

	if ( hid_dev_backup->product == USB_PRODUCT_ID_SIS817_TOUCH || hid_dev_backup->product == USB_PRODUCT_ID_SISF817_TOUCH )	//817 method
	{
		u16 size = (((u16)(buf[64] & 0xff)) << 24) + (((u16)(buf[65] & 0xff)) << 16) + 
			(((u16)(buf[66] & 0xff)) << 8) + (u16)(buf[67] & 0xff);
		int timeout = (((int)(buf[68] & 0xff)) << 24) + (((int)(buf[69] & 0xff)) << 16) + 
			(((int)(buf[70] & 0xff)) << 8) + (int)(buf[71] & 0xff);


		//printk( KERN_INFO "sys_sis_HID_IO %02X", request );
	
		//printk (KERN_INFO "timeout = %d, size %d\n", timeout, size);

		rep_data = kzalloc(size, GFP_KERNEL);
		if (!rep_data)
			return -12;

		if ( copy_from_user( rep_data, (void*)buf, size) ) 
		{
			printk( KERN_INFO "copy_to_user fail\n" );
			//free allocated data
			kfree( rep_data );
			rep_data = NULL;
			return -19;
		}

		rep_ret = usb_interrupt_msg( dev, usbhid->urbout->pipe,
			rep_data, size, &actual_length, timeout );

		if( rep_ret == 0 )
		{
			rep_ret = actual_length;
			if ( copy_to_user( (void*)buf, rep_data, rep_ret ) ) 
			{
				printk( KERN_INFO "copy_to_user fail\n" );
				//free allocated data
				kfree( rep_data );
				rep_data = NULL;
				return -19;
			}
		}
		//printk( KERN_INFO "sis_cdev_write : rep_ret = %d\n", rep_ret );
	}
	else	//816 method
	{

		u8 request = buf[0];
		u8 dir = buf[1];
		u16 value = (((u16)(buf[2] & 0xff)) << 24) + (((u16)(buf[3] & 0xff)) << 16) + (((u16)(buf[4] & 0xff)) << 8) + (u16)(buf[5] & 0xff);
		u16 index = (((u16)(buf[6] & 0xff)) << 24) + (((u16)(buf[7] & 0xff)) << 16) + (((u16)(buf[8] & 0xff)) << 8) + (u16)(buf[9] & 0xff);


		u16 size = (((u16)(buf[29] & 0xff)) << 24) + (((u16)(buf[30] & 0xff)) << 16) + (((u16)(buf[31] & 0xff)) << 8) + (u16)(buf[32] & 0xff);
		int timeout = (((int)(buf[33] & 0xff)) << 24) + (((int)(buf[34] & 0xff)) << 16) + (((int)(buf[35] & 0xff)) << 8) + (int)(buf[36] & 0xff);

		//printk( KERN_INFO "sys_sis_HID_IO %02X", request );
	
		//printk (KERN_INFO "dir = %d, value = %d, index = %d, timeout = %d, buf[33] = %x, buf[34] = %x, buf[35] = %x, buf[36] = %x\n", dir, value, index, timeout, buf[33], buf[34], buf[35], buf[36]);

		rep_data = kzalloc( size , GFP_KERNEL );
		if ( rep_data == 0 ) 
		{
			return -12;
		}

		for( i = 0; i < size; i++)
		{
			rep_data[i] = buf[10+i];
		}

		tmp_data = kzalloc( size + 1, GFP_KERNEL );	//include report id, so size +1

		for( i = 0; i < size; i++ )
		{
			tmp_data[i+1] = rep_data[i];
		}

		buf += 10;

		if( dir & DIR_IN )
		{
			if ( hid_dev_backup->product == USB_PRODUCT_ID_SIS2_TOUCH || hid_dev_backup->product == USB_PRODUCT_ID_NEW_SIS2_TOUCH )
			{
				//20110510, Yuger, for correcting intr data send into interrupt msg(receive, in, endp=2)
				tmp_data[0] = 0x0A;//in
				//printk(KERN_INFO "(INT_IN)tmp_data = %X %X %X, size = %u, actual_length = %d", tmp_data[1], tmp_data[2], tmp_data[3], size, actual_length);

				//816 internal : endpoint 2 is int_in
				rep_ret = usb_interrupt_msg( dev, usb_rcvintpipe( dev, ENDP_02 ), tmp_data, size+1, &actual_length, timeout );
				//yy
				//printk(KERN_INFO "rep_ret = %ld, actual_length = %d", rep_ret, actual_length);
				if( rep_ret == 0 )
				{
					rep_ret = actual_length;
				}

				//20110510, Yuger, for recovering rep_data
				for( i = 0; i < size; i++ )
				{
					rep_data[i] = tmp_data[i+1];
				}
				//yy

				//printk(KERN_INFO "(INT_IN)size = %u, dir = %u, rep_ret = %ld, rep_data = %X %X %X", size, dir, rep_ret, rep_data[0], rep_data[1], rep_data[2]);

				if ( copy_to_user( (void*)buf, rep_data, rep_ret ) ) 
				{
					printk( KERN_INFO "copy_to_user fail\n" );
					//free allocated data
					kfree( rep_data );
					kfree( tmp_data );
					rep_data = NULL;
					tmp_data = NULL;
					return -19;
				}
			}
			else
			{
				//control message
				rep_ret = usb_control_msg( dev, usb_rcvctrlpipe( dev, CTRL ), 
				request, (USB_DIR_IN|USB_TYPE_VENDOR|USB_RECIP_DEVICE), 
				value, index, rep_data, size, timeout );
#if 0
				printk ("(CTRL) size = %d, dir = %d, rep_ret = %ld, rep_data = ", size, dir, rep_ret);
				for (i=0; i<size-1; i++)  
				{
					printk ("%02X ", rep_data[i]);
				}
				if (i == size-1)
				printk ("%02X\n", rep_data[i]);
#endif
				if ( copy_to_user( (void*)buf, rep_data, rep_ret ) ) 
				{
					printk( KERN_INFO "copy_to_user fail\n" );
					//free allocated data
					kfree( rep_data );
					rep_data = NULL;
					return -19;
				}
			}
		}
		else
		{
			if ( hid_dev_backup->product == USB_PRODUCT_ID_SIS2_TOUCH || hid_dev_backup->product == USB_PRODUCT_ID_NEW_SIS2_TOUCH )
			{
				//20110510, Yuger, for correcting intr data send into interrupt msg(send, out, endp=1)
				tmp_data[0] = 0x09;//out

				//816 internal : endpoint 1 is int_out
				rep_ret = usb_interrupt_msg( dev, usb_sndintpipe( dev, ENDP_01 ), tmp_data, size + 1, &actual_length, timeout );

				//just return success or not(no need to return actual_length if succeed)

				//20110510, Yuger, for recovering rep_data
				for( i = 0; i < size; i++ )
				{
					rep_data[i] = tmp_data[i+1];
				}
				//yy

				//printk(KERN_INFO "(INT_OUT)size = %u, actual_length = %d, rep_ret = %ld, rep_data = %x %x %x", size, actual_length, rep_ret, rep_data[0], rep_data[1], rep_data[2]);

				if ( copy_to_user( (void*)buf, rep_data, actual_length-1 ) ) 
				{
					printk( KERN_INFO "copy_to_user fail\n" );
					//free allocated data
					kfree( rep_data );
					kfree( tmp_data );
					rep_data = NULL;
					tmp_data = NULL;
					return -19;
				}
			}
			else
			{
				//control message
				rep_ret = usb_control_msg( dev, usb_sndctrlpipe( dev, CTRL ), 
				request, (USB_DIR_OUT|USB_TYPE_VENDOR|USB_RECIP_DEVICE), 
				value, index, rep_data, 16, timeout );
#if 0
				printk ("(CTRL) size = %d, dir = %d, rep_ret = %ld, rep_data = ", size, dir, rep_ret);
				for (i=0; i<size-1; i++)  
				{
					printk ("%02X ", rep_data[i]);
				}
				if (i == size-1)
					printk ("%02X\n", rep_data[i]);
#endif

				if ( copy_to_user( (void*)buf, rep_data, rep_ret ) ) 
				{
					printk( KERN_INFO "copy_to_user fail\n" );
					//free allocated data
					kfree( rep_data );
					rep_data = NULL;
					return -19;
				}
			}
		}
		//free allocated data
		kfree( tmp_data );
		tmp_data = NULL;
	}
	//free allocated data
	kfree( rep_data );
	rep_data = NULL;

	//printk( KERN_INFO "End of sys_sis_HID_IO\n" );
	return rep_ret;
}


//for ioctl
static const struct file_operations sis_cdev_fops = {
	.owner	= THIS_MODULE,
	.read	= sis_cdev_read,
	.write	= sis_cdev_write,
	.open	= sis_cdev_open,
	.release= sis_cdev_release,
};

//for ioctl
int sis_setup_chardev(struct hid_device *hdev)
{
	
	dev_t dev = MKDEV(sis_char_major, 0);
	int alloc_ret = 0;
	int cdev_err = 0;
	int input_err = 0;
	struct device *class_dev = NULL;
	void *ptr_err;
	
	printk("sis_setup_chardev.\n");
	hid_dev_backup = hdev;
	
	backup_urb = usb_alloc_urb(0, GFP_KERNEL); //0721test
	if (!backup_urb) {
		dev_err(&hdev->dev, "cannot allocate backup_urb\n");
		return -ENOMEM;
	}
/*	
	if (nd == NULL) 
	{
          input_err = -ENOMEM;
          goto error;
	} 
*/
	// dynamic allocate driver handle
	if (hdev->product == USB_PRODUCT_ID_SIS9200_TOUCH)
		alloc_ret = alloc_chrdev_region(&dev, 0, sis_char_devs_count, BRIDGE_DEVICE_NAME);
	else if (hdev->product == USB_PRODUCT_ID_SIS817_TOUCH)
		alloc_ret = alloc_chrdev_region(&dev, 0, sis_char_devs_count, SIS817_DEVICE_NAME);
	else if (hdev->product == USB_PRODUCT_ID_SISF817_TOUCH)
		alloc_ret = alloc_chrdev_region(&dev, 0, sis_char_devs_count, SISF817_DEVICE_NAME);
	else
		alloc_ret = alloc_chrdev_region(&dev, 0, sis_char_devs_count, INTERNAL_DEVICE_NAME);

	if (alloc_ret)
		goto error;
		
	sis_char_major  = MAJOR(dev);
	cdev_init(&sis_char_cdev, &sis_cdev_fops);
	sis_char_cdev.owner = THIS_MODULE;
	cdev_err = cdev_add(&sis_char_cdev, MKDEV(sis_char_major, 0), sis_char_devs_count);
	if (cdev_err) 
		goto error;

	if (hdev->product == USB_PRODUCT_ID_SIS9200_TOUCH)
		printk(KERN_INFO "%s driver(major %d) installed.\n", BRIDGE_DEVICE_NAME, sis_char_major);
	else if (hdev->product == USB_PRODUCT_ID_SIS817_TOUCH)
		printk(KERN_INFO "%s driver(major %d) installed.\n", SIS817_DEVICE_NAME, sis_char_major);
	else if (hdev->product == USB_PRODUCT_ID_SISF817_TOUCH)
		printk(KERN_INFO "%s driver(major %d) installed.\n", SISF817_DEVICE_NAME, sis_char_major);
	else
		printk(KERN_INFO "%s driver(major %d) installed.\n", INTERNAL_DEVICE_NAME, sis_char_major);

	// register class
	if (hdev->product == USB_PRODUCT_ID_SIS9200_TOUCH)
		sis_char_class = class_create(THIS_MODULE, BRIDGE_DEVICE_NAME);
	else if (hdev->product == USB_PRODUCT_ID_SIS817_TOUCH)
		sis_char_class = class_create(THIS_MODULE, SIS817_DEVICE_NAME);
	else if (hdev->product == USB_PRODUCT_ID_SISF817_TOUCH)
		sis_char_class = class_create(THIS_MODULE, SISF817_DEVICE_NAME);
	else
		sis_char_class = class_create(THIS_MODULE, INTERNAL_DEVICE_NAME);

	if(IS_ERR(ptr_err = sis_char_class)) 
	{
		goto err2;
	}

	if (hdev->product == USB_PRODUCT_ID_SIS9200_TOUCH)
		class_dev = device_create(sis_char_class, NULL, MKDEV(sis_char_major, 0), NULL, BRIDGE_DEVICE_NAME);
	else if (hdev->product == USB_PRODUCT_ID_SIS817_TOUCH)
		class_dev = device_create(sis_char_class, NULL, MKDEV(sis_char_major, 0), NULL, SIS817_DEVICE_NAME);
	else if (hdev->product == USB_PRODUCT_ID_SISF817_TOUCH)
		class_dev = device_create(sis_char_class, NULL, MKDEV(sis_char_major, 0), NULL, SISF817_DEVICE_NAME);
	else
		class_dev = device_create(sis_char_class, NULL, MKDEV(sis_char_major, 0), NULL, INTERNAL_DEVICE_NAME);

	if(IS_ERR(ptr_err = class_dev)) 
	{
		goto err;
	}
	
	return 0;
error:
	if (cdev_err == 0)
		cdev_del(&sis_char_cdev);
	if (alloc_ret == 0)
		unregister_chrdev_region(MKDEV(sis_char_major, 0), sis_char_devs_count);
	if(input_err != 0)
	{
		printk("sis_ts_bak error!\n");
	}
err:
	device_destroy(sis_char_class, MKDEV(sis_char_major, 0));
err2:
	class_destroy(sis_char_class);
	return -1;
}
EXPORT_SYMBOL(sis_setup_chardev);

void sis_deinit_chardev(void)
{
	//for ioctl
	dev_t dev;
	printk(KERN_INFO "sis_remove\n");

	dev = MKDEV(sis_char_major, 0);
	cdev_del(&sis_char_cdev);
	unregister_chrdev_region(dev, sis_char_devs_count);
	device_destroy(sis_char_class, MKDEV(sis_char_major, 0));
	class_destroy(sis_char_class);
	usb_kill_urb( backup_urb );
	usb_free_urb( backup_urb );
	backup_urb = NULL;
	hid_dev_backup = NULL;
	
}
EXPORT_SYMBOL(sis_deinit_chardev);

MODULE_DESCRIPTION("SiS 817 Touchscreen Control Driver");
MODULE_LICENSE("GPL");
